#ifndef OSQP_H
#define OSQP_H


/* Types, functions etc required by the OSQP API */
# include "osqp_configure.h"
# include "osqp_api_constants.h"
# include "osqp_api_types.h"
# include "osqp_api_functions.h"

#endif /* ifndef OSQP_H */
